##############################################################################
## 
## FixedEffectCoefficients : vector x in A'Ax = A'b
##
## We define these methods used in lsmr! (duck typing): 
## copyto!, fill!, rmul!, axpy!, norm
##
## Do not define iteration on each fixedeffect since it would conflict with eltype
##
##############################################################################

struct FixedEffectCoefficients{T}
	x::Vector{<:AbstractVector{T}}
end

eltype(xs::FixedEffectCoefficients{T}) where {T} = T
length(xs::FixedEffectCoefficients) = sum(length(x) for x in xs.x)
norm(xs::FixedEffectCoefficients) = sqrt(sum(sum(abs2, x) for x in xs.x))

function fill!(xs::FixedEffectCoefficients, α::Number)
	for x in xs.x
		fill!(x, α)
	end
	return xs
end

function rmul!(xs::FixedEffectCoefficients, α::Number)
	for x in xs.x
		rmul!(x, α)
	end
	return xs
end

function copyto!(xs1::FixedEffectCoefficients, xs2::FixedEffectCoefficients)
	for (x1, x2) in zip(xs1.x, xs2.x)
		copyto!(x1, x2)
	end
	return xs1
end

function axpy!(α::Number, xs1::FixedEffectCoefficients, xs2::FixedEffectCoefficients)
	for (x1, x2) in zip(xs1.x, xs2.x)
		axpy!(α, x1, x2)
	end
	return xs2
end

##############################################################################
## 
## FixedEffectLSMR
##
## A is the model matrix of categorical variables
## normalized by diag(1/a1, ..., 1/aN) (Jacobi preconditoner)
##
## We define these methods used in lsmr! (duck typing):
## mul!
##
##############################################################################

struct FixedEffectLSMR{T} <: AbstractFixedEffectMatrix{T}
	fes::Vector{<:FixedEffect}
	scales::Vector{<:AbstractVector}
	caches::Vector{<:AbstractVector}
	xs::FixedEffectCoefficients{T}
	v::FixedEffectCoefficients{T}
	h::FixedEffectCoefficients{T}
	hbar::FixedEffectCoefficients{T}
	u::AbstractVector{T}
	r::AbstractVector{T}
	sqrtw::AbstractVector{T}
end

adjoint(fem::FixedEffectLSMR) = Adjoint(fem)

function size(fem::FixedEffectLSMR, dim::Integer)
	(dim == 1) ? length(fem.fes[1].refs) : (dim == 2) ? sum(fe.n for fe in fem.fes) : 1
end

function mul!(y::AbstractVector, fem::FixedEffectLSMR, 
			  fecoefs::FixedEffectCoefficients, α::Number, β::Number)
	rmul!(y, β)
	for (fecoef, fe, cache) in zip(fecoefs.x, fem.fes, fem.caches)
		demean!(y, fecoef, fe.refs, α, cache)
	end
	return y
end

function demean!(y::AbstractVector, fecoef::AbstractVector, refs::AbstractVector, 
				 α::Number, cache::AbstractVector)
	@simd ivdep for i in eachindex(y)
		@inbounds y[i] += fecoef[refs[i]] * α * cache[i]
	end
end

function mul!(fecoefs::FixedEffectCoefficients, Cfem::Adjoint{T, FixedEffectLSMR{T}},
				y::AbstractVector, α::Number, β::Number) where {T}
	fem = adjoint(Cfem)
	rmul!(fecoefs, β)
	for (fecoef, fe, cache) in zip(fecoefs.x, fem.fes, fem.caches)
		mean!(fecoef, fe.refs, y, α, cache)
	end
	return fecoefs
end

function mean!(fecoef::AbstractVector, refs::AbstractVector, y::AbstractVector, 
			   α::Number, cache::AbstractVector)
	@simd ivdep for i in eachindex(y)
		@inbounds fecoef[refs[i]] += y[i] * α * cache[i]
	end
end

function solve!(feM::FixedEffectLSMR, r::AbstractVector; 
				tol::Real = 1e-8, maxiter::Integer = 100_000)
	fill!(feM.xs, 0.0)
	copyto!(feM.u, r)
	x, ch = lsmr!(feM.xs, feM, feM.u, feM.v, feM.h, feM.hbar; 
				  atol = tol, btol = tol, conlim = 1e8, maxiter = maxiter)
	return div(ch.mvps, 2), ch.isconverged
end


##############################################################################
##
## Implement AbstractFixedEffectMatrix interface
##
##############################################################################\

function AbstractFixedEffectMatrix{T}(fes::Vector{<:FixedEffect}, sqrtw::AbstractVector, ::Type{Val{:lsmr}}) where {T}
	sqrtw = convert(AbstractVector{T}, sqrtw)
	scales = [scale!(zeros(T, fe.n), fe.refs, fe.interaction, sqrtw) for fe in fes]
	caches = [cache!(zeros(T, length(sqrtw)), fe.refs, fe.interaction, scale, sqrtw) for (fe, scale) in zip(fes, scales)]
	xs = FixedEffectCoefficients([zeros(T, fe.n) for fe in fes])
	v = FixedEffectCoefficients([zeros(T, fe.n) for fe in fes])
	h = FixedEffectCoefficients([zeros(T, fe.n) for fe in fes])
	hbar = FixedEffectCoefficients([zeros(T, fe.n) for fe in fes])
	u = zeros(T, length(sqrtw))
	r = zeros(T, length(sqrtw))
	return FixedEffectLSMR(fes, scales, caches, xs, v, h, hbar, u, r, sqrtw)
end

function scale!(fecoef::AbstractVector, refs::AbstractVector, interaction::AbstractVector, sqrtw::AbstractVector)
	@inbounds @simd for i in eachindex(refs)
		fecoef[refs[i]] += abs2(interaction[i] * sqrtw[i])
	end
	fecoef .= 1 ./ sqrt.(fecoef)
end

function cache!(y::AbstractVector, refs::AbstractVector, interaction::AbstractVector, fecoef::AbstractVector, sqrtw::AbstractVector)
	@inbounds @simd for i in eachindex(y)
		y[i] = fecoef[refs[i]] * interaction[i] * sqrtw[i]
	end
	return y
end

function solve_residuals!(r::AbstractVector, feM::FixedEffectLSMR; kwargs...)
	copyto!(feM.r, r)
	feM.r .*= feM.sqrtw   
	iterations, converged = solve!(feM, feM.r; kwargs...)
	mul!(feM.r, feM, feM.xs, -1.0, 1.0)
	feM.r ./=  feM.sqrtw
	copyto!(r, feM.r)
	return r, iterations, converged
end

function solve_residuals!(X::AbstractMatrix, feM::AbstractFixedEffectMatrix; kwargs...)
	iterations = Int[]
	convergeds = Bool[]
	for x in eachcol(X)
		_, iteration, converged = solve_residuals!(x, feM; kwargs...)
		push!(iterations, iteration)
		push!(convergeds, converged)
	end
	return X, iterations, convergeds
end

function solve_coefficients!(r::AbstractVector, feM::FixedEffectLSMR; kwargs...)
	copyto!(feM.r, r)
	feM.r .*= feM.sqrtw
	iterations, converged = solve!(feM, feM.r; kwargs...)
	for (x, scale) in zip(feM.xs.x, feM.scales)
		x .*=  scale
	end
	xs = Vector{eltype(r)}[x for x in feM.xs.x]
	full(normalize!(xs, feM.fes; kwargs...), feM.fes), iterations, converged
end

